# type: ignore

import pytest
from unittest.mock import MagicMock
import sys


class CAN:
    def __init__(self, *args, **kwargs):
        self.sent_message = None

    @property
    def state(self):
        return "ERROR_ACTIVE"

    def send(self, message):
        self.sent_message = message


class Message:
    is_mock = True

    def __init__(self, id: int, data: bytes):
        self.id = id
        self.data = data


# A fixture that runs once per test session
@pytest.fixture(scope="session", autouse=True)
def mock_hardware_modules():
    # Mock the hardware-specific modules
    board_mock = MagicMock()
    board_mock.CAN_RX = "CAN_RX"
    board_mock.CAN_TX = "CAN_TX"
    sys.modules["board"] = board_mock

    sys.modules["digitalio"] = MagicMock()

    canio = MagicMock()
    canio.Message = Message
    canio.CAN = CAN
    sys.modules["canio"] = canio

    # Optional: Cleanup after the test session is done
    yield  # This allows the test session to run with the mocks in place

    # cleanup
    del sys.modules["board"]
    del sys.modules["digitalio"]
    del sys.modules["canio"]



#!/usr/bin/env python3
"""
 Inspect icu messages on can bus

 Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""
import struct

import can

import rox_icu.can_protocol as canp
from rox_icu.can_utils import get_can_bus
from rox_icu.utils import run_main


def process_message(msg: can.Message):

    node_id, opcode = canp.split_message_id(msg.arbitration_id)
    try:
        decoded = canp.decode_message(msg.arbitration_id, msg.data)
        if isinstance(decoded, canp.HeartbeatMessage):
            print(".", end="", flush=True)
        else:
            print(f"\n{node_id=}, {opcode=}, {decoded}")
    except (KeyError, struct.error):
        pass


def main():

    print("Starting inspector")
    with get_can_bus() as bus:
        try:
            while True:
                msg = bus.recv()
                if msg is not None:
                    process_message(msg)

        except KeyboardInterrupt:
            print("Inspector stopped")


if __name__ == "__main__":
    run_main(main)



# test ICU class with a mock device

import asyncio

import pytest

from rox_icu.core import ICU
from rox_icu.mock import ICUMock

NODE_ID = 1


@pytest.mark.asyncio
async def test_loopback():
    """set a pin on the mock and await for the event on the ICU"""

    mock = ICUMock(NODE_ID)
    mock_task = asyncio.create_task(mock.main())

    icu = ICU(NODE_ID)
    assert icu.node_id == NODE_ID

    try:

        await asyncio.wait_for(icu.start(), timeout=1.0)
        await asyncio.sleep(0.2)  # increment counter

        hb = icu.last_heartbeat
        assert hb is not None
        assert hb.counter > 0

        icu.check_alive()
        pin = icu.pins[0]
        pin.state = True  # set command over can bus

        await asyncio.wait_for(pin.high_event.wait(), timeout=1.0)
        assert mock.io_state == 0x01
        assert not pin.high_event.is_set()

    except asyncio.TimeoutError:
        assert False, "Timeout waiting for high event"

    finally:

        # stop the mock
        try:
            mock_task.cancel()
            await mock_task
        except asyncio.CancelledError:
            pass

        await icu.stop()



#!/usr/bin/env python3
"""
 supporting functions for working with can

 Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""

import os
import logging
from can.interfaces.udp_multicast import UdpMulticastBus
from can.interfaces.socketcan import SocketcanBus


def is_ci_environment() -> bool:
    """Check if the code is running in a CI environment"""
    return os.getenv("CI") == "true"


def get_can_bus() -> UdpMulticastBus | SocketcanBus:
    """Get a CAN bus instance, using environment variables
    CAN_CHANNEL and CAN_INTERFACE for configuration or a multicast bus in CI"""

    if is_ci_environment():
        return UdpMulticastBus("224.0.0.1", interface="udp_multicast")

    channel = os.getenv("CAN_CHANNEL")
    if channel is None:
        raise ValueError(
            "Missing CAN_CHANNEL environment variable, set to can0 or similar"
        )

    interface = os.getenv("CAN_INTERFACE", "socketcan")
    logging.info(f"Using CAN interface: {interface}, channel: {channel}")
    if interface == "udp_multicast":
        # note: will always receive own messages
        return UdpMulticastBus(channel, interface=interface)

    return SocketcanBus(
        channel=channel, interface=interface, receive_own_messages=False
    )



#!/usr/bin/env python3
"""
monitor devices on can bus.
Listens to heartbeat, displays status in curses.

Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""

from __future__ import annotations

import curses
import signal
import struct
from dataclasses import dataclass

import can

from rox_icu import can_protocol as canp
from rox_icu.can_utils import get_can_bus


HB_OPCODE, HB_BYTEDEF = canp.get_opcode_and_bytedef(canp.HeartbeatMessage)

devices: dict[int, Device] = {}
should_exit = False


@dataclass
class Device:
    node_id: int
    is_icu: bool = False
    last_heartbeat: canp.HeartbeatMessage | None = None

    def get_display_data(self) -> str:
        """Get formatted string for display, handling unknown devices and message changes"""
        if not self.is_icu or self.last_heartbeat is None:
            return f"| {self.node_id:<6} | {'--':<7} | {'--':<7} | {'--':<7} | {'--':<8} | {'--':<6}| {'--':<7} |"

        hb = self.last_heartbeat
        # Convert io_state to binary string, pad to 8 bits
        io_state_bin = format(getattr(hb, "io_state", 0), "08b")

        # Use getattr with default to handle possible changes in heartbeat message
        dev_type = getattr(hb, "device_type", "--")
        err1 = getattr(hb, "error_max1", "--")
        err2 = getattr(hb, "error_max2", "--")
        errors = getattr(hb, "errors", "--")
        counter = getattr(hb, "counter", "--")

        return f"| {self.node_id:<6} | {dev_type:<7} | {err1:<7} | {err2:<7} | {io_state_bin:<8} | 0x{errors:<4x}| {counter:<7} |"


def handle_msg(msg: can.Message) -> None:
    # check if the message is a heartbeat
    node_id, opcode = canp.split_message_id(msg.arbitration_id)

    if opcode == HB_OPCODE:
        try:
            hb_msg = canp.decode_message(msg.arbitration_id, msg.data)
            if not isinstance(hb_msg, canp.HeartbeatMessage):
                raise ValueError("Invalid heartbeat message")
            if node_id not in devices:
                devices[node_id] = Device(node_id, is_icu=True, last_heartbeat=hb_msg)
            else:
                devices[node_id].last_heartbeat = hb_msg
        except struct.error:  # wrong number of bytes
            if node_id not in devices:
                devices[node_id] = Device(node_id, is_icu=False)


def signal_handler(signum, frame) -> None:
    """Handle keyboard interrupt"""
    global should_exit  # pylint: disable=global-statement
    should_exit = True


def draw_table(pad: curses.window, screen: curses.window) -> None:
    """Draw the table using a pad"""
    height, width = screen.getmaxyx()
    pad.erase()

    # Draw header
    header = f"| {'NodeID':<6} | {'DevType':<7} | {'ErrMax1':<7} | {'ErrMax2':<7} | {'IO State':<8} | {'Error':<5} | {'Counter':<7} |"
    pad.addstr(0, 0, header)
    pad.addstr(1, 0, "-" * len(header))

    # Draw data
    for idx, (_, device) in enumerate(sorted(devices.items()), start=2):
        row = device.get_display_data()
        pad.addstr(idx, 0, row)

    # Calculate visible area
    visible_rows = height - 1
    visible_cols = width - 1

    # Copy the pad contents to the screen
    pad.refresh(0, 0, 0, 0, visible_rows, visible_cols)


def main(stdscr: curses.window) -> None:
    # Set up curses
    curses.curs_set(0)  # Hide the cursor
    curses.use_default_colors()  # Use terminal's default colors

    # Create a pad larger than the screen
    _, width = stdscr.getmaxyx()
    pad = curses.newpad(1000, width)  # Support up to 1000 rows

    # Set up signal handler for Ctrl+C
    signal.signal(signal.SIGINT, signal_handler)

    try:
        with get_can_bus() as bus:
            while not should_exit:
                # Check for messages
                msg = bus.recv(timeout=0.1)  # 100ms timeout
                if msg is not None:
                    handle_msg(msg)

                # Update display
                draw_table(pad, stdscr)

    except Exception as e:
        curses.endwin()
        raise e


if __name__ == "__main__":
    try:
        curses.wrapper(main)
    except KeyboardInterrupt:
        print("\nExiting gracefully...")



#!/usr/bin/env python3
"""
rox_icu CLI
"""


import can
import click

import rox_icu.can_protocol as canp
from rox_icu import __version__
from rox_icu.utils import run_main
from rox_icu.can_utils import get_can_bus


@click.group()
@click.version_option(version=__version__)
def cli() -> None:
    pass  # pragma: no cover


@cli.command()
def info() -> None:
    """Print package info"""
    print(__version__)


@cli.command()
@click.option("--node-id", default=1, help="device ID")
@click.option("--sim-inputs", is_flag=True, help="simulate inputs")
def mock(
    node_id,
    sim_inputs,
):
    """Mock ICU device on CAN bus"""
    from .mock import ICUMock

    def main() -> None:
        dev = ICUMock(node_id, simulate_inputs=sim_inputs)
        dev.start()

    run_main(main)


@cli.command()
def monitor():
    """Monitor ICU devices on CAN bus"""
    import curses

    import rox_icu.monitor as icu_monitor

    curses.wrapper(icu_monitor.main)


@cli.command()
def inspect():
    """Inspect ICU messages on CAN bus (ignoring heartbeat)"""
    from .inspector import main

    run_main(main)


@cli.command()
@click.argument("hex_input")
@click.option("--node-id", default=1, help="device ID")
def output(hex_input: str, node_id: int):
    """Set output state, provide hex value"""
    try:
        state = int(hex_input, 16)

    except ValueError:
        print("Invalid hex input. Please provide a valid hexadecimal value.")
        return

    # for now, simply put packet on the bus, without checking if device is alive

    print(f"Setting ICU_{node_id} output state to {state:08b}")

    with get_can_bus() as bus:
        arb_id, data = canp.encode_message(canp.IOSetMessage(state), node_id)
        msg = can.Message(arbitration_id=arb_id, data=data, is_extended_id=False)
        bus.send(msg)


if __name__ == "__main__":
    cli()  # pragma: no cover



"""pin definitions


Feather M4 CAN pins  and ECU mapping (for reference)
Symbols:
 - "=" means same as on Feather
 - "x" means not connected
 - "!" means rempapped
-----------------------------------

= board.A0 board.D14 (PA02)  <- VCC10V measure 10V reference
! board.A1 board.D15 (PA05)  -> MAX_CRCEN
! board.A2 board.D16 (PB08)  <- MAX1_NFAULT
! board.A3 board.D17 (PB09)  <- MAX2_NFAULT
! board.A4 board.D18 (PA04)  -> MAX_ENABLE
! board.A5 board.D19 (PA06)  <- MAX1_NVDDOK
! board.BATTERY board.VOLTAGE_MONITOR (PB00) <- A0
= board.BOOST_ENABLE (PB13)
= board.CAN_RX (PB15)
= board.CAN_STANDBY (PB12)
= board.CAN_TX (PB14)
! board.D0 board.RX (PB17)  -> MAX_SYNCH
! board.D1 board.TX (PB16)  <- MAX_NREADY
! board.D10 (PA20)          <-> MAX2_D2
! board.D11 (PA21)          <-> MAX2_D3
! board.D12 (PA22)          <-> MAX2_D4
= board.D13 board.LED (PA23)
= board.D23 board.MISO (PB22)
= board.D24 board.MOSI (PB23)
= board.D25 board.SCK (PA17)
! board.D4 (PA14)           -> MAX_CS
! board.D5 (PA16)            <-> MAX1_D3
! board.D6 (PA18)           <-> MAX1_D4
! board.D9 (PA19)          <-> MAX2_D1
= board.NEOPIXEL (PB02)
x board.NEOPIXEL_POWER (PB03)
! board.SCL (PA13)          <-> MAX1_D2
! board.SDA (PA12)         <-> MAX1_D1

"""

from microcontroller import pin  # type: ignore
from digitalio import DigitalInOut, Direction

import neopixel
import canio
import busio
import max14906 as maxio


class Pins:
    """pin definitions"""

    NEOPIXEL = pin.PB02

    # status leds
    LED1 = pin.PA23
    LED2 = pin.PA27
    LED = pin.PA23  # consistent with Feather M4

    VCC10V = pin.PA02
    A0 = pin.PB00
    A1 = pin.PB01

    # enable 5V
    BOOST_ENABLE = pin.PB13

    # can
    CAN_STANDBY = pin.PB12
    CAN_TX = pin.PB14
    CAN_RX = pin.PB15

    # spi
    SCK = pin.PA17
    MOSI = pin.PB23
    MISO = pin.PB22

    # max 14906 pins
    MAX_CRCEN = pin.PA05
    MAX_SYNCH = pin.PB17
    MAX_NREADY = pin.PB16
    MAX_CS = pin.PA14
    MAX_ENABLE = pin.PA04
    MAX1_NFAULT = pin.PB08
    MAX2_NFAULT = pin.PB09
    MAX1_NVDDOK = pin.PA06
    MAX2_NVDDOK = pin.PA07

    # max d_pins
    MAX1_D1 = pin.PA12
    MAX1_D2 = pin.PA13
    MAX1_D3 = pin.PA16
    MAX1_D4 = pin.PA18
    MAX2_D1 = pin.PA19
    MAX2_D2 = pin.PA20
    MAX2_D3 = pin.PA21
    MAX2_D4 = pin.PA22


# ----- object instances -----

# leds
led1 = DigitalInOut(Pins.LED1)
led1.switch_to_output()

led2 = DigitalInOut(Pins.LED2)
led2.switch_to_output()

rgb_led = neopixel.NeoPixel(Pins.NEOPIXEL, 1)
# set dim green on module import
rgb_led.brightness = 0.01
rgb_led.fill((0, 255, 0))

# 5V supply, enable by default
boost_enable = DigitalInOut(Pins.BOOST_ENABLE)
boost_enable.switch_to_output()
boost_enable.value = True

# can interface
_can_standby = DigitalInOut(Pins.CAN_STANDBY)
_can_standby.switch_to_output(False)

can = canio.CAN(rx=Pins.CAN_RX, tx=Pins.CAN_TX, baudrate=500_000, auto_restart=True)


# spi interface
spi = busio.SPI(Pins.SCK, Pins.MOSI, Pins.MISO)

# lock the spi bus
_spi_lock = False
for retry in range(10):
    _spi_lock = spi.try_lock()
    if _spi_lock:
        print(f"spi locked on attempt {retry}")
        break

if not _spi_lock:
    raise RuntimeError("SPI lock failed")

del _spi_lock

spi.configure(baudrate=500_000, phase=0, polarity=0)


# max14906 pins
max_cs = DigitalInOut(Pins.MAX_CS)
max_cs.switch_to_output(True)

max_synch = DigitalInOut(Pins.MAX_SYNCH)
max_synch.switch_to_output(True)

max_nready = DigitalInOut(Pins.MAX_NREADY)

max_crcen = DigitalInOut(Pins.MAX_CRCEN)
max_crcen.switch_to_output(False)

max_enable = DigitalInOut(Pins.MAX_ENABLE)
max_enable.switch_to_output(False)

max1_nfault = DigitalInOut(Pins.MAX1_NFAULT)
max2_nfault = DigitalInOut(Pins.MAX2_NFAULT)


# max chips
_dpins = [
    DigitalInOut(pin)
    for pin in [Pins.MAX1_D1, Pins.MAX1_D2, Pins.MAX1_D3, Pins.MAX1_D4]
]
max1 = maxio.Max14906(spi, max_cs, _dpins, chip_address=0)
del _dpins

_dpins = [
    DigitalInOut(pin)
    for pin in [Pins.MAX2_D1, Pins.MAX2_D2, Pins.MAX2_D3, Pins.MAX2_D4]
]
max2 = maxio.Max14906(spi, max_cs, _dpins, chip_address=1)
del _dpins

# -------------convenience class for D pins----------------


class D_Pin:
    """wrapper around DigitalInOut and corresponding maxio pin"""

    def __init__(self, max_chip: maxio.Max14906, pin_nr: int):
        self.chip = max_chip
        self.pin_nr = pin_nr

    def switch_to_output(self, value: bool) -> None:
        self.chip.switch_to_output(self.pin_nr, value)

    def switch_to_input(self) -> None:
        self.chip.switch_to_input(self.pin_nr)

    @property
    def value(self) -> bool:
        return self.chip.d_pins[self.pin_nr].value

    @value.setter
    def value(self, value: bool):
        self.chip.d_pins[self.pin_nr].value = value

    @property
    def direction(self) -> Direction:
        return self.chip.d_pins[self.pin_nr].direction

    @direction.setter
    def direction(self, direction: Direction):
        self.chip.d_pins[self.pin_nr].direction = direction

    def __repr__(self):
        dir_sign = "input" if self.direction == Direction.INPUT else "output"
        return f"D_Pin({self.chip.chip_address}, {self.pin_nr},{dir_sign})"


D_PINS: list[D_Pin] = []
for max_chip in [max2, max1]:
    for pin_nr in range(4):
        D_PINS.append(D_Pin(max_chip, pin_nr))



#!/usr/bin/env python3
"""
Simulate ICU remote IO device.

Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""

import asyncio
import struct
import time
import gc
import can_protocol as canp
import canio
from icu_board import D_PINS, can, led1, max1, max2, max_enable
from micropython import const
from digitalio import Direction

NODE_ID = const(0x01)
device_errors = 0

# -------------- initialisation ----------------
print("----------Remote IO sumulator-------------")
print(f"Node ID: {NODE_ID}")
print(f"Can protocol version: {canp.VERSION}")

# intialize system
max_enable.value = True  # enable in- and outputs

# TODO: get pin configurations for outputs from settings.toml
# see https://docs.circuitpython.org/en/latest/docs/environment.html
# for pin_nr in [BUTTON, SENSOR]:
#     D_PINS[pin_nr].switch_to_input()

# show state of all d pins
for pin in D_PINS:
    print(pin)


def get_io_state() -> int:
    io_state = 0

    for bit, pin in enumerate(D_PINS):
        io_state |= pin.value << bit
    return io_state


def set_io_state(state: int) -> None:
    """set output pins based on state (byte)"""
    for bit, pin in enumerate(D_PINS):
        if pin.direction == Direction.OUTPUT:
            pin.value = (state >> bit) & 0x01


async def read_inputs() -> None:
    """read inputs and send can message on change"""

    global device_errors

    io_state = get_io_state()
    prev_io_state = io_state

    opcode, byte_def = canp.get_opcode_and_bytedef(canp.IOStateMessage)
    msg_id = canp.generate_message_id(NODE_ID, opcode)
    print(f"IOStateMessage ID: {msg_id:x}")

    # Timing variables
    loop_count = 0
    max_loop_time = 0.0
    total_loop_time = 0.0
    loop_start = time.monotonic_ns()

    while True:
        # get state of all pins
        io_state = get_io_state()

        if io_state != prev_io_state:
            msg = canio.Message(id=msg_id, data=struct.pack(byte_def, io_state))
            can.send(msg)

        prev_io_state = io_state

        # Calculate loop time in milliseconds
        loop_time = (time.monotonic_ns() - loop_start) / 1_000_000  # Convert ns to ms

        # Update statistics
        max_loop_time = max(max_loop_time, loop_time)
        total_loop_time += loop_time
        loop_count += 1

        # Print statistics every 100 loops
        if loop_count % 100 == 0:
            avg_loop_time = total_loop_time / 100
            print(
                f"Loop timing - Avg: {avg_loop_time:.3f}ms, Max: {max_loop_time:.3f}ms"
            )

            # set error if required
            if max_loop_time > 10:
                print("*****************Long looptime********************")
                device_errors = device_errors | 0x01

            # Reset statistics
            max_loop_time = 0.0
            total_loop_time = 0.0

        gc.collect()

        loop_start = time.monotonic_ns()
        await asyncio.sleep(0)


async def heartbeat_loop() -> None:
    counter = 0

    opcode, byte_def = canp.get_opcode_and_bytedef(canp.HeartbeatMessage)

    msg_id = canp.generate_message_id(NODE_ID, opcode)
    print(f"HeartbeatMessage ID: {msg_id:x}")

    while True:
        msg = canp.HeartbeatMessage(
            device_type=1,
            error_max1=max1.get_global_error(),
            error_max2=max2.get_global_error(),
            io_state=get_io_state(),
            errors=device_errors,
            counter=counter & 0xFF,
        )
        # print(msg)
        can_msg = canio.Message(id=msg_id, data=struct.pack(byte_def, *msg))
        can.send(can_msg)

        led1.value = not led1.value

        counter += 1

        await asyncio.sleep(0.1)


async def receive_can_message() -> None:
    listener = can.listen(timeout=0)
    while True:
        if listener.in_waiting():
            msg = listener.receive()
            if msg:
                if canp.get_node_id(msg.id) == NODE_ID:
                    if isinstance(msg, canio.RemoteTransmissionRequest):
                        print(f"RTR message {msg.id:x}")
                    else:
                        decoded_msg = canp.decode_message(msg.id, msg.data)
                        print(f"Received message: {decoded_msg}")

                        if isinstance(decoded_msg, canp.IOStateMessage):
                            set_io_state(decoded_msg.io_state)
                            print(f"IO state set to: {decoded_msg.io_state}")

        await asyncio.sleep(0)  # Yield control to other tasks


# -------------------test function --------------------------------


async def toggle_outputs() -> None:
    """testing function to toggle outputs"""
    pin = D_PINS[0]

    while True:
        pin.value = not pin.value
        await asyncio.sleep(0.5)


# ---------------------main---------------------------------------------


async def main() -> None:
    await asyncio.gather(
        read_inputs(),
        heartbeat_loop(),
        toggle_outputs(),
        receive_can_message(),
    )


asyncio.run(main())



# SPDX-FileCopyrightText: 2020 anecdata for Adafruit Industries
# SPDX-FileCopyrightText: 2021 Neradoc for Adafruit Industries
# SPDX-FileCopyrightText: 2021-2023 Kattni Rembor for Adafruit Industries
# SPDX-FileCopyrightText: 2023 Dan Halbert for Adafruit Industries
#
# SPDX-License-Identifier: MIT

# CircuitPython Essentials Pin Map Script
# This script generates a list of all the pins available on the board,
# along with their corresponding aliases.

import microcontroller
import board

# Initialize a list to hold pin mappings
mapped_pins = []

# Iterate over every attribute in the microcontroller.pin module
for pin_name in dir(microcontroller.pin):
    pin_obj = getattr(microcontroller.pin, pin_name)
    # Check if the attribute is an instance of a Pin
    if isinstance(pin_obj, microcontroller.Pin):
        # Temporarily store aliases for this pin
        pin_aliases = []
        # Check each attribute in the board module
        for alias in dir(board):
            if getattr(board, alias) is pin_obj:
                # If the board attribute is the same as the pin, add it to the aliases
                pin_aliases.append(f"board.{alias}")
        # If there are any aliases, append the original GPIO name and the aliases to the list
        if pin_aliases:
            pin_aliases.append(f"({str(pin_name)})")
            mapped_pins.append(" ".join(pin_aliases))

# Print each pin and its aliases, sorted for readability
for pin_info in sorted(mapped_pins):
    print(pin_info)



#!/usr/bin/env python3
"""
 test output pins

 Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""
import asyncio
import os

from rox_icu.core import ICU
from rox_icu.utils import run_main_async

NODE_ID = 1
INTERFACE = os.getenv("ICU_INTERFACE", "vcan0")
print(f"Using interface: {INTERFACE}")


async def main() -> None:

    icu = ICU(NODE_ID, INTERFACE)
    await icu.start()

    for i in range(32):

        icu.io_state = i
        await asyncio.sleep(0.1)

    icu.io_state = 0

    await icu.stop()
    print("Done")


if __name__ == "__main__":
    run_main_async(main())



#!/usr/bin/env python3
"""
ICU CAN driver

Copyright (c) 2024 ROX Automation
"""
from __future__ import annotations

import asyncio
import logging
import threading
import time
from typing import Literal, Optional

import can
from can.interfaces.socketcan import SocketcanBus
from can.interfaces.udp_multicast import UdpMulticastBus

from rox_icu import can_protocol as canp
from rox_icu.can_utils import get_can_bus

# message timeout in seconds
MESSAGE_TIMEOUT = 1.0  # message expiration time & can timeout
HEARTBEAT_TIMEOUT = 0.5  # heartbeat specific timeout


class DeviceError(Exception):
    """ICU device error"""


class HeartbeatError(Exception):
    """No heartbeat error"""


class AutoClearEvent(asyncio.Event):
    """Event that automatically clears itself after all waiters are done"""

    def __init__(self) -> None:
        super().__init__()
        self._active_waiters: int = 0

    async def wait(self) -> Literal[True]:
        self._active_waiters += 1
        try:
            return await super().wait()
        finally:
            self._active_waiters -= 1
            if self._active_waiters == 0:
                self.clear()


class Pin:
    """Represents a single I/O pin with its state and events"""

    def __init__(
        self, number: int, is_input: bool = False, parent: ICU | None = None
    ) -> None:
        self._parent = parent
        self._number: int = number
        self._is_input: bool = is_input
        self._state: bool = False
        self._previous_state: bool = False
        self.high_event: AutoClearEvent = AutoClearEvent()
        self.low_event: AutoClearEvent = AutoClearEvent()
        self.change_event: AutoClearEvent = AutoClearEvent()

    @property
    def number(self) -> int:
        return self._number

    @property
    def is_input(self) -> bool:
        return self._is_input

    @property
    def state(self) -> bool:
        return self._state

    @state.setter
    def state(self, new_state: bool) -> None:
        if self._is_input:
            raise ValueError("pin is read-only")

        if self._parent is None:
            raise RuntimeError("parent is not set")

        self._parent.io_state = self._parent.io_state & ~(1 << self._number) | (
            new_state << self._number
        )

    def _update(self, new_state: bool) -> None:
        """update pin state, set events, used by ICU class"""

        if new_state == self._state:
            return

        self._previous_state = self._state
        self._state = new_state

        # Set appropriate events
        self.change_event.set()
        if new_state:
            self.high_event.set()
        else:
            self.low_event.set()


class ICU:
    """ICU CAN driver"""

    def __init__(
        self,
        node_id: int,
        can_bus: UdpMulticastBus | SocketcanBus | None = None,
    ) -> None:
        self._log = logging.getLogger(f"icu.{node_id}")
        self._node_id = node_id

        self._bus = can_bus or get_can_bus()

        self._receive_thread: Optional[threading.Thread] = None
        self._msg_task: Optional[asyncio.Task] = None
        self._msg_queue: asyncio.Queue = asyncio.Queue()

        self._last_heartbeat: Optional[canp.HeartbeatMessage] = None
        self._last_heartbeat_time: float = 0
        self._heartbeat_event = asyncio.Event()

        self._ignored_messages: set = set()
        self._running = True

        self._io_state: int = 0

        self.pins = [Pin(i, parent=self) for i in range(8)]

    @property
    def node_id(self) -> int:
        """Node ID"""
        return self._node_id

    @property
    def io_state(self) -> int:
        """Get IO state"""
        return self._io_state

    @io_state.setter
    def io_state(self, state: int) -> None:
        """Set output state, provide a byte for all 8 outputs"""

        self._log.debug(f"> {self._node_id=} {state=}")

        self.check_alive()

        arb_id, msg_data = canp.encode_message(
            canp.IOSetMessage(state), node_id=self._node_id
        )

        self._bus.send(
            can.Message(arbitration_id=arb_id, data=msg_data, is_extended_id=False)
        )

    def check_alive(self) -> None:
        """Check if device is alive, raise an exception if not"""
        if self._last_heartbeat is None:
            raise HeartbeatError("Error: No heartbeat message received.")

        if time.time() - self._last_heartbeat_time > HEARTBEAT_TIMEOUT:
            raise HeartbeatError("Error: Heartbeat message timeout.")

    async def wait_for_heartbeat(self, timeout: float = 1.0) -> None:
        """Wait for heartbeat message"""
        self._heartbeat_event.clear()
        self._log.debug("waiting for heartbeat")
        try:
            await asyncio.wait_for(self._heartbeat_event.wait(), timeout)
        except asyncio.TimeoutError as e:
            raise HeartbeatError("Timeout waiting for heartbeat") from e

    @property
    def last_heartbeat(self) -> canp.HeartbeatMessage | None:
        """Get last heartbeat message"""
        return self._last_heartbeat

    async def start(self) -> None:
        """Start driver"""
        self._log.info(
            f"Starting. node_id={self._node_id}, bus={self._bus.channel_info}"
        )

        loop = asyncio.get_running_loop()
        self._receive_thread = threading.Thread(
            target=self._can_reader_thread, args=(loop,), daemon=True
        )
        self._receive_thread.start()

        self._msg_task = asyncio.create_task(self._message_handler())

        # Wait for first heartbeat
        self._log.info("waiting for first heartbeat")
        await self.wait_for_heartbeat()

        self._log.info("started")

    async def stop(self) -> None:
        """Stop driver"""
        self._log.info("stopping driver")
        self._running = False

        if self._msg_task is not None:
            self._msg_task.cancel()
            try:
                await self._msg_task
            except asyncio.CancelledError:
                pass

        if self._receive_thread is not None:
            self._receive_thread.join(timeout=1.0)

        if hasattr(self, "_bus"):
            self._bus.shutdown()

        self._log.info("stopped")

    def _can_reader_thread(self, loop: asyncio.AbstractEventLoop) -> None:
        """Receive CAN messages, filter and put them into the queue"""
        timeout_warned = False

        while self._running:
            try:
                msg = self._bus.recv(MESSAGE_TIMEOUT)
                if not msg:
                    if not timeout_warned:
                        self._log.warning("can timeout")
                        timeout_warned = True
                    continue
                else:
                    if timeout_warned:
                        self._log.info("message flow restored")
                    timeout_warned = False

                node_id, opcode = canp.split_message_id(msg.arbitration_id)

                # Ignore messages that aren't for this node
                if node_id != self._node_id:
                    continue

                # Ignore messages that were requested to be ignored
                if opcode in self._ignored_messages:
                    continue

                # RTR messages are requests for data
                if msg.is_remote_frame:
                    self._log.warning("RTR message received")
                    continue

                self._log.debug(f"< {node_id=} {opcode=}")

                if self._running:  # Check again before queueing
                    asyncio.run_coroutine_threadsafe(
                        self._msg_queue.put((msg.arbitration_id, bytes(msg.data))), loop
                    )

            except Exception as e:
                if self._running:  # Only log if we're still meant to be running
                    self._log.error(f"Error in CAN reader thread: {e}")
                break

        self._log.debug("CAN reader thread stopped")

    def _uptate_io_state(self, io_state: int) -> None:
        """Update IO state"""
        self._io_state = io_state

        for i in range(8):
            # pylint: disable=protected-access
            self.pins[i]._update(bool(io_state & (1 << i)))

    async def _message_handler(self) -> None:
        """Handle received messages"""
        while self._running:
            try:
                arb_id, data = await self._msg_queue.get()
                self._msg_queue.task_done()

                msg = canp.decode_message(arb_id, data)

                if isinstance(msg, canp.HeartbeatMessage):
                    self._last_heartbeat = msg
                    self._last_heartbeat_time = time.time()
                    self._uptate_io_state(self._last_heartbeat.io_state)
                    self._heartbeat_event.set()
                    self._log.debug(f"heartbeat: {self._last_heartbeat}")

                elif isinstance(msg, canp.IOStateMessage):
                    self._uptate_io_state(msg.io_state)

            except asyncio.CancelledError:
                break
            except Exception as e:
                self._log.error(f"Error processing message: {e}")

        self._log.debug("Message handler stopped")



# example of time measurement
import gc
import time

for _ in range(50):
    t_start = time.monotonic_ns()
    time.sleep(0.001)
    t_end = time.monotonic_ns()
    t_elapsed = (t_end - t_start) / 1e6
    error = abs(t_elapsed - 1)

    print(f"elapsed: {t_elapsed:.3f} ms (error: {error:.1%})")
    gc.collect()



#!/usr/bin/env python3
"""
Support functions

Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""

import asyncio
import logging
import os
from typing import Any, Coroutine, Callable, Union, TypeVar

import coloredlogs

LOG_FORMAT = "%(asctime)s [%(name)s] %(filename)s:%(lineno)d - %(message)s"
TIME_FORMAT = "%H:%M:%S.%f"

# Type variable for the return type of the callable/coroutine
T = TypeVar("T")

# Define a type alias for our supported function types
MainFunction = Union[
    Callable[[], T],  # Regular synchronous function
    Coroutine[Any, Any, T],  # Coroutine
]


def setup_logging() -> None:
    """Setup logging"""
    loglevel = os.environ.get("LOGLEVEL", "INFO").upper()
    coloredlogs.install(level=loglevel, fmt=LOG_FORMAT, datefmt=TIME_FORMAT)
    logging.info(f"Log level set to {loglevel}")


def get_root_exception(exc: BaseException) -> BaseException:
    """Traverse the exception chain to find the root cause."""
    if isinstance(exc, ExceptionGroup):
        # If it's an ExceptionGroup, recursively check its exceptions
        for e in exc.exceptions:
            return get_root_exception(e)
    while exc.__cause__ is not None:
        exc = exc.__cause__
    return exc


def run_main(func: Callable) -> None:
    setup_logging()

    try:
        func()
    except KeyboardInterrupt:
        logging.info("Process interrupted by user")
    except ExceptionGroup as group:
        root_exc = get_root_exception(group)
        logging.error(
            f"Root cause: {type(root_exc).__name__}: {str(root_exc)}",
            exc_info=True,
        )

    except Exception as e:
        logging.error(e, exc_info=True)


def run_main_async(
    coro: Coroutine[Any, Any, None],
    silence_loggers: list[str] | None = None,
) -> None:
    """convenience function to avoid code duplication"""

    setup_logging()

    if silence_loggers:
        for logger in silence_loggers:
            logging.info(f"Silencing logger: {logger}")
            logging.getLogger(logger).setLevel(logging.WARNING)

    try:
        asyncio.run(coro)
    except KeyboardInterrupt:
        print("KeyboardInterrupt")
    except ExceptionGroup as group:
        root_exc = get_root_exception(group)
        logging.error(
            f"Root cause: {type(root_exc).__name__}: {str(root_exc)}", exc_info=True
        )
    except asyncio.CancelledError:
        logging.error("Cancelled")
    except Exception as e:
        root_exc = get_root_exception(e)
        logging.error(
            f"Root cause: {type(root_exc).__name__}: {str(root_exc)}", exc_info=True
        )
